#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from scipy.stats import norm, ks_2samp, ranksums, f
import statsmodels.api as sm

#################################################################
# rdrandinf observed statistics and asymptotic p-values
#################################################################

def rdrandinf_model(Y, D, statistic, pvalue=False, kweights=None, endogtr=None, delta=None):
    
    try: n = len(D)
    except: n = 1
    n1 = np.sum(D)
    n0 = n - n1

    try: lenY = len(Y)
    except: lenY = 1
    Y = np.array(Y).reshape(lenY,-1)
   
    if statistic == 'ttest' or statistic == 'diffmeans':
        if np.all(kweights == 1):
            Y1 = Y[D == 1]
            Y0 = Y[D == 0]
            M1 = np.mean(Y1, axis=0)
            M0 = np.mean(Y0, axis=0)
            stat = M1 - M0
            if pvalue:
                V1 = np.mean((Y1 - np.mean(Y1, axis=0)) ** 2, axis=0) / (n1 - 1)
                V0 = np.mean((Y0 - np.mean(Y0, axis=0)) ** 2, axis=0) / (n0 - 1)
                se = np.sqrt(V1 + V0)
                t_stat = (M1 - M0) / se
                asy_pval = 2 * norm.cdf(-np.abs(t_stat))
                if delta is not None:
                    asy_power = 1 - norm.cdf(1.96 - delta / se) + norm.cdf(-1.96 - delta / se)
                else:
                    asy_power = np.nan
        else:
            stat = np.zeros(Y.shape[1])
            asy_pval = np.zeros(Y.shape[1])
            for k in range(Y.shape[1]):
                lm_aux = sm.WLS(Y[:, k], sm.add_constant(D), weights=kweights).fit()
                stat[k] = lm_aux.params[1]
                if pvalue:
                    se = np.sqrt(lm_aux.cov_params().loc[1, 1])
                    t_stat = stat[k] / se
                    asy_pval[k] = 2 * norm.cdf(-np.abs(t_stat))
                    if delta is not None:
                        asy_power = 1 - norm.cdf(1.96 - delta / se) + norm.cdf(-1.96 - delta / se)
                    else:
                        asy_power = np.nan

    elif statistic == 'ksmirnov':
        stat = np.zeros(Y.shape[1])
        asy_pval = np.zeros(Y.shape[1])
        for k in range(Y.shape[1]):
            aux_ks = ks_2samp(Y[D == 0, k], Y[D == 1, k])
            stat[k] = aux_ks.statistic
            if pvalue:
                asy_pval[k] = aux_ks.pvalue
                asy_power = np.nan

    elif statistic == 'ranksum':
        stat = np.zeros(Y.shape[1])
        asy_pval = np.zeros(Y.shape[1])
        for k in range(Y.shape[1]):
            T_stat, _ = ranksums(Y[D == 0, k], Y[D == 1, k])
            stat[k] = T_stat
        if pvalue:
            asy_pval = 2 * norm.cdf(-np.abs(stat))
            sigma = np.std(Y, axis=0, ddof = 1)
            if delta is not None:
                asy_power = norm.cdf(np.sqrt(3 * n0 * n1 / ((n0 + n1 + 1) * np.pi)) * delta / sigma - 1.96)
            else:
                asy_power = np.nan

    elif statistic == 'all':
        stat1 = np.mean(Y[D == 1,0]) - np.mean(Y[D == 0,0])
        aux_ks = ks_2samp(Y[D == 0,0], Y[D == 1,0])
        stat2 = aux_ks.statistic
        aux_rs = ranksums(Y[D == 0,0], Y[D == 1,0])
        stat3 = aux_rs.statistic
        stat = np.array([stat1, stat2, stat3])
        if pvalue:
            Y1 = Y[D == 1,0]
            Y0 = Y[D == 0,0]
            V1 = np.mean((Y1 - np.mean(Y1, axis=0)) ** 2, axis=0) / (n1 - 1)
            V0 = np.mean((Y0 - np.mean(Y0, axis=0)) ** 2, axis=0) / (n0 - 1)
            se1 = np.sqrt(V1 + V0)
            t_stat = stat1 / se1
            asy_pval1 = 2 * norm.cdf(-np.abs(t_stat))
            asy_pval2 = aux_ks.pvalue
            asy_pval3 = 2 * norm.cdf(-np.abs(stat3))
            asy_pval = np.array([asy_pval1, asy_pval2, asy_pval3])
            if delta is not None:
                asy_power1 = 1 - norm.cdf(1.96 - delta / se1) + norm.cdf(-1.96 - delta / se1)
                asy_power2 = np.nan
                sigma = np.std(Y[:,0], axis=0, ddof = 1)
                asy_power3 = norm.cdf(np.sqrt(3 * n0 * n1 / ((n0 + n1 + 1) * np.pi)) * delta / sigma - 1.96)
                asy_power = np.array([asy_power1, asy_power2, asy_power3])
            else:
                asy_power = np.array([np.nan, np.nan, np.nan])
    
    elif statistic == 'ar':
        stat = np.mean(Y[D == 1], axis=0) - np.mean(Y[D == 0], axis=0)
        if pvalue:
            se = np.sqrt(np.var(Y[D == 1], axis=0) / n1 + np.var(Y[D == 0], axis=0) / n0)
            t_stat = stat / se
            asy_pval = 2 * norm.cdf(-np.abs(t_stat))
            if delta is not None:
                asy_power = 1 - norm.cdf(1.96 - delta / se) + norm.cdf(-1.96 - delta / se)
            else:
                asy_power = np.nan
    
    elif statistic == 'wald':
        fs = sm.OLS(endogtr, sm.add_constant(D)).fit()
        rf = sm.OLS(Y, sm.add_constant(D)).fit()
        stat = rf.params[1] / fs.params[1]
        if pvalue:
            ehat = Y - np.mean(Y) - stat * (endogtr - np.mean(endogtr))
            ehat2 = ehat ** 2
            se = np.sqrt((np.mean(ehat2) * np.var(D)) / (n * np.cov(D, endogtr) ** 2))
            t_stat = stat / se
            asy_pval = 2 * norm.cdf(-np.abs(t_stat))
            if delta is not None:
                asy_power = 1 - norm.cdf(1.96 - delta / se) + norm.cdf(-1.96 - delta / se)
            else:
                asy_power = np.nan

    if pvalue:
        output = {'statistic': stat, 'p_value': asy_pval, 'asy_power': asy_power}
    else:
        output = {'statistic': stat}

    return output

#################################################################
# Hotelling's T2 statistic
#################################################################

def hotelT2(X, D):
    n = len(D)
    n1 = np.sum(D)
    n0 = n - n1
    p = X.shape[1]

    X1 = X[D == 1, :]
    X0 = X[D == 0, :]
    X1bar = np.mean(X1, axis=0)
    X0bar = np.mean(X0, axis=0)
    S1 = np.cov(X1.T)
    S0 = np.cov(X0.T)
    Spool = (S1 * (n1 - 1) + S0 * (n0 - 1)) / (n - 2)
    SpoolInv = np.linalg.inv(Spool)

    T2 = (n0 * n1 / n) * np.dot(np.dot((X1bar - X0bar).T, SpoolInv), (X1bar - X0bar))
    Fstat = ((n - p - 1) / ((n - 2) * p)) * T2
    pval = 1 - f.cdf(Fstat, p, n - 1 - p)

    output = {'statistic': float(T2), 'Fstat': float(Fstat), 'p.value': float(pval)}

    return output

#################################################################
# Find window increments
#################################################################

def findwobs(wobs, nwin, posl, posr, R, dups):
    N = len(R)
    Nc = np.sum(R < 0)
    Nt = np.sum(R >= 0)
    mpoints_l = len(np.unique(R[:Nc]))
    mpoints_r = len(np.unique(R[Nc:N]))
    mpoints_max = max(mpoints_l, mpoints_r)
    nwin_mp = min(nwin, mpoints_max)
    poslold = posl
    posrold = posr

    win = 1
    wlist_left = []
    poslist_left = []
    wlist_right = []
    poslist_right = []

    while (win <= nwin_mp) and (wobs < max(posl, Nt - (posr - Nc - 1))):

        poslold = posl
        posrold = posr

        while (dups[posl-1] < wobs) and (np.sum(R[posl-1] <= R[(posl-1):poslold]) < wobs) and (posl > 1):
            posl = max(posl - dups[posl-1], 1)

        while (dups[posr-1] < wobs) and (np.sum(R[(posrold-1):posr] <= R[posr-1]) < wobs) and (posr < N):
            posr = min(posr + dups[posr-1], N)

        wlength_left = R[posl-1]
        wlength_right = R[posr-1]

        wlist_left.append(wlength_left)
        poslist_left.append(posl)
        wlist_right.append(wlength_right)
        poslist_right.append(posr)

        posl = max(posl - dups[posl-1], 1)
        posr = min(posr + dups[posr-1], N)

        win += 1

    output = {
        'posl': posl,
        'posr': posr,
        'wlength_left': wlength_left,
        'wlength_right': wlength_right,
        'wlist_left': wlist_left,
        'wlist_right': wlist_right,
        'poslist_left': poslist_left,
        'poslist_right': poslist_right,
    }

    return output

#################################################################
# Find symmetric window increments
#################################################################

def findwobs_sym(wobs, nwin, posl, posr, R, dups):
    N = len(R)
    Nc = np.sum(R < 0)
    Nt = np.sum(R >= 0)
    poslold = posl
    posrold = posr
    wlist = []
    win = 1

    while (win <= nwin) and (wobs < min(posl, Nt - (posr - Nc - 1))):

        poslold = posl
        posrold = posr

        while (dups[posl-1] < wobs) and (np.sum(R[posl-1] <= R[(posl-1):poslold]) < wobs):
            posl = max(posl - dups[posl-1], 1)

        while (dups[posr-1] < wobs) and (np.sum(R[(posrold-1):posr] <= R[posr-1]) < wobs):
            posr = min(posr + dups[posr-1], N)

        if abs(R[posl-1]) < R[posr-1]:
            posl = Nc + 1 - np.sum(-R[posr-1] <= R[:Nc])

        if abs(R[posl-1]) > R[posr-1]:
            posr = np.sum(R[Nc:N] <= abs(R[posl-1])) + Nc

        wlength = max(-R[posl-1], R[posr-1])
        wlist.append(wlength)

        posl = max(posl - dups[posl-1], 1)
        posr = min(posr + dups[posr-1], N)
        win += 1

    return np.array(wlist)

#################################################################
# Find CI
#################################################################

def find_CI(pvals, alpha, tlist):
    if np.all(pvals >= alpha):
        CI = np.array([[tlist[0], tlist[-1]]])
    elif np.all(pvals < alpha):
        CI = np.full((1, 2), np.nan)
    else:
        whichvec = np.where(pvals >= alpha)[0]
        index_l = np.min(whichvec)
        index_r = np.max(whichvec)
        indexmat = np.array([[index_l, index_r]])

        whichvec_cut = whichvec.copy()
        dif = np.diff(whichvec_cut)
        while np.all(dif == 1) is False:
            cut = np.min(np.where(dif != 1))
            auxvec = whichvec_cut[:cut + 1]
            indexmat = np.vstack((indexmat, [np.min(auxvec), np.max(auxvec)]))
            whichvec_cut = whichvec_cut[cut + 1:]

            dif = np.diff(whichvec_cut)

        if indexmat.shape[0] > 1:
            indexmat = indexmat[1:, :]
            indexmat = np.vstack((indexmat, [np.min(whichvec_cut), np.max(whichvec_cut)]))
        CI = np.array([[tlist[i] for i in indexmat[0]]])

    return CI

#################################################################
# Find window length - DEPRECATED: for backward compatibility
#################################################################

def wlength(R, D, num):
    X1 = np.sort(np.abs(R[D == 1]))
    X0 = np.sort(np.abs(R[D == 0]))
    m = min(len(X1), len(X0))
    if num > m:
        num = m
    xt = X1[num - 1]
    xc = X0[num - 1]
    minw = max(xc, xt)
    return minw

#################################################################
# Find default step - DEPRECATED: for backward compatibility
#################################################################

def findstep(R, D, obsmin, obsstep, times):
    S = []
    for i in range(1, times + 1):
        U = wlength(R, D, obsmin + obsstep * i)
        L = wlength(R, D, obsmin + obsstep * (i - 1))
        Snext = U - L
        S.append(Snext)
    step = max(S)
    return step





###############################################################################
# rdlocrand: illustration file
# !version 1.0 21-Jun-2023
# Authors: Matias Cattaneo, Ricardo Masini, Rocio Titiunik, Gonzalo Vazquez-Bare
###############################################################################

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

########## To run from the source ##################################################
# (in terminal) pip uninstall rdlocrand 
import sys
sys.path.insert(0, '/Users/rmasini/Library/CloudStorage/Dropbox/rdlocrand/python/rdlocrand/src')
from rdwinselect import rdwinselect
from rdrandinf import rdrandinf
from rdsensitivity import rdsensitivity
from rdrbounds import rdrbounds
###################################################################################

###############################################################################
## Load data
###############################################################################

data = pd.read_csv("/Users/rmasini/Library/CloudStorage/Dropbox/rdlocrand/R/rdlocrand_senate.csv")

# Select predetermined covariates to be used for window selector

X = data[["presdemvoteshlag1", "population", "demvoteshlag1", "demvoteshlag2",
          "demwinprv1", "demwinprv2", "dopen", "dmidterm", "dpresdem"]]

X.loc[:,'population'] = X['population'].div(1000000)

# Assign names to the covariates

colnames_X = ["DemPres Vote", "Population", "DemSen Vote t-1", "DemSen Vote t-2",
              "DemSen Win t-1", "DemSen Win t-2", "Open", "Midterm", "DemPres"]

X.columns = colnames_X

# Running variable and outcome variable

R = data["demmv"].values
Y = data["demvoteshfor2"].values
D = 1*(R >= 0)

##############################################################################
# rdwinselect
##############################################################################

# Deprecated default options (Stata Journal 2016)
tmp = rdwinselect(R, X, obsstep=2)

# Window selection with default options
tmp = rdwinselect(R, X)

# Window selection with default options and symmetric windows
tmp = rdwinselect(R, X, wasymmetric=True)

# Window selection with specified window length and increments (replicate CFT)
tmp = rdwinselect(R, X, wmin=0.5, wstep=0.125, reps=10000)

# Window selection using large sample approximation and plotting p-values
tmp = rdwinselect(R, X, wmin=0.5, wstep=0.125, approx=True, nwindows = 80, quietly=True, plot=True)

###############################################################################
## rdrandinf
###############################################################################

# Randomization inference using recommended window
tmp = rdrandinf(Y, R, wl=-0.75, wr=0.75)

# Randomization inference using recommended window, all statistics
tmp = rdrandinf(Y, R, wl=-0.75, wr=0.75, statistic='all')

# Randomization inference using recommended window using rdwinselect
tmp = rdrandinf(Y, R, statistic='all', covariates=X, wmin=0.5, wstep=0.125, rdwreps=10000)

# Randomization inference using recommended window, linear adjustment
tmp = rdrandinf(Y, R, wl=-0.75, wr=0.75, statistic='all', p=1)

# Randomization inference under interference
tmp = rdrandinf(Y, R, wl=-0.75, wr=0.75, interfci=0.05)

###############################################################################
# rdsensitivity
###############################################################################

tmp = rdsensitivity(Y, R, wlist=np.arange(0.75, 10.25, 0.25), tlist=np.arange(0, 21, 1))

# Replicate contour plot
xaxis = tmp['wlist']
yaxis = tmp['tlist']
zvalues = tmp['results']
plt.contourf(xaxis, yaxis, zvalues, levels=np.arange(0, 1.01, 0.01), cmap='gray')
plt.xlabel('window')
plt.ylabel('treatment effect')
plt.colorbar(label='p-value')
plt.title('Sensitivity Analysis')
plt.show()

# Obtain 95 percent confidence interval for window [-0.75 ; 0.75]
tmp = rdsensitivity(Y, R, wlist=np.arange(0.75, 2.25, 0.25), tlist=np.arange(0, 21, 1), ci=[-0.75, 0.75])
confidence_interval = tmp['ci']
print(confidence_interval)

# rdsensitivity to calculate CI from within rdrandinf
tmp = rdrandinf(Y, R, wl=-0.75, wr=0.75, ci=[0.05] + list(np.arange(3, 21, 1)))


###############################################################################
# rdrbounds
###############################################################################

tmp = rdrbounds(Y, R, expgamma=[1.5, 2, 3], wlist=[0.5, 0.75, 1], reps=1000)
lower_bound = tmp['lower.bound']
print(lower_bound)
upper_bound = tmp['upper.bound']
print(upper_bound)

# Bernoulli and fixed margins p-values
tmp = rdrbounds(Y, R, expgamma=[1.5, 2, 3], wlist=[0.5, 0.75, 1], reps=1000, fmpval=True)
lower_bound = tmp['lower.bound']
print(lower_bound)
upper_bound = tmp['upper.bound']
print(upper_bound)


###############################################################################
# rdrandinf with eval options
###############################################################################

ii = (R >= -0.75) & (R <= 0.75) & (~np.isnan(Y)) & (~np.isnan(R))
m0 = np.mean(R[ii & (D == 0)], axis=0)
m1 = np.mean(R[ii & (D == 1)], axis=0)
tmp = rdrandinf(Y, R, wl=-0.75, wr=0.75, p=1, evall=m0, evalr=m1)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
import warnings
from scipy.stats import norm, binom_test
import statsmodels.api as sm
import matplotlib.pyplot as plt
from rdlocrand.rdlocrand_fun import findwobs, findwobs_sym, findstep, hotelT2, rdrandinf_model

def rdwinselect(R, X=None, cutoff=0, obsmin=None, wmin=None, wobs=None, wstep=None,
                wasymmetric=False, wmasspoints=False, dropmissing=False, nwindows=10,
                statistic='diffmeans', p=0, evalat='cutoff', kernel='uniform',
                approx=False, level=0.15, reps=1000, seed=666, plot=False, quietly=False,
                obsstep=None):
    
    """
    Window selection for RD designs under local randomization

    rdwinselect implements the window-selection procedure based on balance tests for RD designs
    under local randomization. Specifically, it constructs a sequence of nested windows around
    the RD cutoff and reports binomial tests for the running variable runvar and covariate
    balance tests for covariates (if specified). The recommended window is the largest window
    around the cutoff such that the minimum p-value of the balance test is larger than a
    prespecified level for all nested (smaller) windows. By default, the p-values are calculated
    using randomization inference methods.

    Parameters:
    ----------
    R : array-like
        A vector containing the values of the running variable.
    X : array-like, optional
        The matrix of covariates to be used in the balancing tests. The matrix is optional but
        the recommended window is only provided when at least one covariate is specified.
        This should be a matrix of size n x k where n is the total sample size and k is the number
        of covariates.
    cutoff : float, optional
        The RD cutoff (default is 0).
    obsmin : int, optional
        The minimum number of observations above and below the cutoff in the smallest window.
        Default is 10.
    wmin : float, optional
        The smallest window to be used.
    wobs : int, optional
        The number of observations to be added at each side of the cutoff at each step.
        Default is 5.
    wasymmetric : bool, optional
        Allows for asymmetric windows around the cutoff when wobs is specified.
    wmasspoints : bool, optional
        Specifies that the running variable is discrete and each masspoint should be used as a window.
    wstep : float, optional
        The increment in window length.
    nwindows : int, optional
        The number of windows to be used. Default is 10.
    dropmissing : bool, optional
        Drop rows with missing values in covariates when calculating windows.
    statistic : str, optional
        The statistic to be used in the balance tests. Allowed options are 'diffmeans'
        (difference in means statistic), 'ksmirnov' (Kolmogorov-Smirnov statistic),
        'ranksum' (Wilcoxon-Mann-Whitney standardized statistic), and 'hotelling'
        (Hotelling's T-squared statistic). Default option is 'diffmeans'. The statistic
        'ttest' is equivalent to 'diffmeans' and included for backward compatibility.
    p : int, optional
        The order of the polynomial for outcome adjustment model (for covariates). Default is 0.
    evalat : str, optional
        Specifies the point at which the adjusted variable is evaluated. Allowed options are
        'cutoff' and 'means'. Default is 'cutoff'.
    kernel : str, optional
        Specifies the type of kernel to use as a weighting scheme. Allowed kernel types are
        'uniform' (uniform kernel), 'triangular' (triangular kernel), and 'epan' (Epanechnikov kernel).
        Default is 'uniform'.
    approx : bool, optional
        Forces the command to conduct the covariate balance tests using a large-sample approximation
        instead of finite-sample exact randomization inference methods.
    level : float, optional
        The minimum accepted value of the p-value from the covariate balance tests. Default is 0.15.
    reps : int, optional
        Number of replications. Default is 1000.
    seed : int, optional
        The seed to be used for the randomization tests.
    plot : bool, optional
        Draws a scatter plot of the minimum p-value from the covariate balance test against window length.
    quietly : bool, optional
        Suppress output.
    obsstep : int, optional
        The minimum number of observations to be added on each side of the cutoff for the sequence
        of fixed-increment nested windows. This option is deprecated and only included for backward
        compatibility.

    Returns:
    -------
    window : float or None
        Recommended window (None if covariates are not specified).
    wlist : list
        List of window lengths.
    results : DataFrame
        Table including window lengths, minimum p-value in each window, corresponding number of
        the variable with the minimum p-value (i.e., column of covariate matrix), Binomial test
        p-value, and sample sizes to the left and right of the cutoff in each window.
    summary : dict
        Summary statistics.

    Examples:
    ---------
    X = np.random.randn(100, 2)
    R = X[0] + X[1] + np.random.randn(100)

    # Window selection adding 5 observations at each step
    # Note: low number of replications to speed up process.
    tmp = rdwinselect(R, X, obsmin=10, wobs=5, reps=500)

    # Window selection setting initial window and step
    # The user should increase the number of replications.
    tmp = rdwinselect(R, X, wmin=0.5, wstep=0.125, reps=500)

    # Window selection with approximate (large sample) inference and p-value plot
    tmp = rdwinselect(R, X, wmin=0.5, wstep=0.125, approx=True, nwin=80, quietly=True, plot=True)
    """
    
    ###############################################################################
    # Parameters and error checking
    ###############################################################################
    
    if cutoff <= np.min(R) or cutoff >= np.max(R):
        raise ValueError('Cutoff must be within the range of the running variable')
    if p < 0:
        raise ValueError('p must be a positive integer')
    if p > 0 and approx and (statistic != 'ttest' and statistic != 'diffmeans'):
        raise ValueError('approximate and p > 1 can only be combined with diffmeans')
    valid_statistics = ['diffmeans', 'ttest', 'ksmirnov', 'ranksum', 'hotelling']
    if statistic not in valid_statistics:
        raise ValueError(f'{statistic} not a valid statistic')
    valid_evalat = ['cutoff', 'means']
    if evalat not in valid_evalat:
        raise ValueError('evalat only admits means or cutoff')
    valid_kernels = ['uniform', 'triangular', 'epan']
    if kernel not in valid_kernels:
        raise ValueError(f'{kernel} not a valid kernel')
    if kernel != 'uniform' and evalat != 'cutoff':
        raise ValueError('kernel can only be combined with evalat(cutoff)')
    if kernel != 'uniform' and statistic != 'ttest' and statistic != 'diffmeans':
        raise ValueError('kernel only allowed for diffmeans')
    if obsmin is not None and wmin is not None:
        raise ValueError('cannot set both obsmin and wmin')
    if wobs is not None and wstep is not None:
        raise ValueError('cannot set both wobs and wstep')
    if wmasspoints:
        if obsmin is not None:
            raise ValueError('obsmin not allowed with wmasspoints')
        if wmin is not None:
            raise ValueError('wmin not allowed with wmasspoints')
        if wobs is not None:
            raise ValueError('wobs not allowed with wmasspoints')
        if wstep is not None:
            raise ValueError('wstep not allowed with wmasspoints')
    
    Rc = pd.DataFrame(np.array(R - cutoff), columns =['Rc'])
    D = pd.DataFrame(1*np.array(R >= cutoff), columns =['D'])

    if X is not None:
        X = pd.DataFrame(X)
        colnames_X = X.columns.values
        data = pd.concat([Rc, D, X], axis = 1)
        if not dropmissing: data = data.dropna(subset=['Rc', 'D'])
        else: data = data.dropna()
        data = data.sort_values('Rc')
        X = data.drop(['Rc', 'D'],axis=1).values
    else:
        colnames_X = None
        data = pd.concat([Rc, D], axis = 1)
        data = data.dropna()
        data = data.sort_values('Rc')
    
    Rc = data['Rc'].values
    D = data['D'].values
    
    if seed > 0:
        np.random.seed(seed)
    elif seed != -1:
        raise ValueError('Seed has to be a positive integer or -1 for system seed')
    
    testing_method = 'rdrandinf' if not approx else 'approximate'
    
    n = len(Rc)
    n1 = np.sum(D)
    n0 = n - n1
    Rc_df = pd.DataFrame({'Rc': Rc})
    count = Rc_df.groupby('Rc').size().to_frame(name = "count")
    dups = pd.merge(Rc_df, count, on='Rc', sort = True)['count'].values
    
    if np.max(dups) > 1:
        print('Mass points detected in running variable')
        print('You may use wmasspoints option for constructing windows at each mass point')
        mp_left = np.unique(Rc[D == 0])
        mp_right = np.unique(Rc[D == 1])
        if wmasspoints:
            nmax = min(max(len(mp_left), len(mp_right)), nwindows)
            wlist = np.empty((2, nmax))
    
    ###############################################################################
    # Define initial window
    ###############################################################################
    
    if wmin is None:
        posl = n0
        posr = n0 + 1
        
        if obsmin is None:
            obsmin = 10
        if wmasspoints:
            obsmin = 1
            wasymmetric = True
        if obsstep is not None:
            wmin = findwobs_sym(obsmin, 1, posl, posr, Rc, dups)
        if wasymmetric:
            tmp = findwobs(obsmin, 1, posl, posr, Rc, dups)
            wmin_left = tmp['wlength_left']
            posmin_left = tmp['poslist_left'][0]-1
            wmin_right = tmp['wlength_right']
            posmin_right = tmp['poslist_right'][0]-1
        else:
            wmin_right = findwobs_sym(obsmin, 1, posl, posr, Rc, dups)
            wmin_left = -wmin_right
    
    else:
        if np.isscalar(wmin):
            wmin_right = wmin
            wmin_left = -wmin
            posmin_right = n0 + np.sum(np.logical_and(Rc <= wmin, Rc >= 0))
            posmin_left = n0 - np.sum(np.logical_and(Rc < 0, Rc >= -wmin)) + 1
        elif len(wmin) == 2:
            wmin_left = wmin[0]
            wmin_right = wmin[1]
            posmin_right = n0 + np.sum(np.logical_and(Rc <= wmin_right, Rc >= 0))
            posmin_left = n0 - np.sum(np.logical_and(Rc < 0, Rc >= wmin_left)) + 1
        else:
            raise ValueError('wmin option incorrectly specified')
    
    ###############################################################################
    # Define window list
    ###############################################################################
    
    if obsstep is not None:
        warnings.warn('obsstep included for backward compatibility only.\nThe use of wstep and wobs is recommended.')
        wstep = findstep(Rc, D, obsmin, obsstep, 10)
        wlist_right = np.linspace(wmin[0], wmin[0] + wstep * (nwindows - 1), num=nwindows)
        wlist_left = None
    elif wstep is not None:
        wmax_left = max(wmin_left - wstep * (nwindows - 1), min(Rc))
        wmax_right = min(wmin_right + wstep * (nwindows - 1), max(Rc))
        wlist_left = np.sort(np.arange(wmax_left, wmin_left+wstep, step=wstep))[::-1]
        wlist_right = np.arange(wmin_right, wmax_right + wstep, step=wstep)
    else:
        if wobs is None:
            wobs = 5
        if wmasspoints:
            wobs = 1
        posl = max(n0 - np.sum(np.logical_and(Rc < 0, Rc >= wmin_left)), 1)
        posr = min(n0 + 1 + np.sum(np.logical_and(Rc >= 0, Rc <= wmin_right)), n)
        if wasymmetric:
            tmp = findwobs(wobs, nwindows - 1, posl, posr, Rc, dups)
            wlist_left = np.concatenate(([wmin_left], tmp['wlist_left']))
            poslist_left = np.concatenate(([posmin_left], np.array(tmp['poslist_left'])-1))
            wlist_right = np.concatenate(([wmin_right], tmp['wlist_right']))
            poslist_right = np.concatenate(([posmin_right], np.array(tmp['poslist_right'])-1))
        else:
            wlist = findwobs_sym(wobs, nwindows - 1, posl, posr, Rc, dups)
            wlist_right = np.concatenate((wmin_right, wlist))
            wlist_left = np.concatenate((wmin_left, wlist))
    
    nmax = min(nwindows, len(wlist_right))
    if nmax < nwindows:
        print()
        warnings.warn('Not enough observations to calculate all windows. '
                      'Consider changing wmin(), wobs(), or wstep().')
        
    ###############################################################################
    # Summary statistics
    ###############################################################################

    table_sumstats = np.empty((5, 2))
    table_sumstats[0, :] = [n0, n1]

    qq0 = np.round(np.quantile(np.abs(Rc[D == 0]), q=[0.01, 0.05, 0.1, 0.2], interpolation='lower'), 5)
    qq1 = np.round(np.quantile(Rc[D == 1], q=[0.01, 0.05, 0.1, 0.2], interpolation='lower'), 5)

    n0_q1 = np.sum((Rc >= -qq0[0]) & (Rc < 0))
    n0_q2 = np.sum((Rc >= -qq0[1]) & (Rc < 0))
    n0_q3 = np.sum((Rc >= -qq0[2]) & (Rc < 0))
    n0_q4 = np.sum((Rc >= -qq0[3]) & (Rc < 0))
    n1_q1 = np.sum((Rc <= qq1[0]) & (Rc >= 0))
    n1_q2 = np.sum((Rc <= qq1[1]) & (Rc >= 0))
    n1_q3 = np.sum((Rc <= qq1[2]) & (Rc >= 0))
    n1_q4 = np.sum((Rc <= qq1[3]) & (Rc >= 0))

    table_sumstats[1, :] = [n0_q1, n1_q1]
    table_sumstats[2, :] = [n0_q2, n1_q2]
    table_sumstats[3, :] = [n0_q3, n1_q3]
    table_sumstats[4, :] = [n0_q4, n1_q4]

    ###############################################################################
    ## Display upper-right panel
    ###############################################################################

    if not quietly:
        print('\n')
        print('Window selection for RD under local randomization')
        print('\n')
        print(f"{'Number of obs':18}= {n:14}")
        print(f"{'Order of poly':18}= {p:14}")
        print(f"{'Kernel type':18}= {kernel:>14}")
        print(f"{'Reps':18}= {reps:14}")
        print(f"{'Testing method':18}= {testing_method:>14}")
        print(f"{'Balance test':18}= {statistic:>14}")
        print('\n')

    ###############################################################################
    ## Display upper left panel
    ###############################################################################
    
    if not quietly:
        print(format("Cutoff c = ", '10s'), format(f'{cutoff:.3f}', '8s'), format("Left of c", '12s'),format("Right of c", '12s'))
        print(format("Number of obs", '19s'), format(f'{n0:6.0f}', '12s'), format(f'{n1:6.0f}', '12s'))
        print(format("1st percentile", '19s'), format(f'{n0_q1:6.0f}', '12s'), format(f'{n1_q1:6.0f}', '12s'))
        print(format("5th percentile", '19s'), format(f'{n0_q2:6.0f}', '12s'), format(f'{n1_q2:6.0f}', '12s'))
        print(format("10th percentile", '19s'), format(f'{n0_q3:6.0f}', '12s'), format(f'{n1_q3:6.0f}', '12s'))
        print(format("20th percentile", '19s'), format(f'{n0_q4:6.0f}', '12s'), format(f'{n1_q4:6.0f}', '12s'))
        print('\n')

    ###############################################################################
    # Balance tests
    ###############################################################################

    table_rdw = np.full((nmax, 7),np.nan)

    ## Being main panel display

    if not quietly:
        print("=" * 80)
        print(f"{'Window':^18}{'p-value':>11}{'Var. name':>16}{'Bin.test':>12}{'Obs<c':>11}{'Obs>=c':>11}")
        print("=" * 80)

    for j in range(nmax):
        if wasymmetric and wstep is None and obsstep is None:
            wlower = wlist_left[j]
            wupper = wlist_right[j]

            position_l = poslist_left[j]
            position_r = poslist_right[j]

            ww = (Rc >= Rc[position_l]) & (Rc <= Rc[position_r])

        else:
            wupper = wlist_right[j]
            wlower = -wupper
            if wlist_left is not None: wlist_left[j] = wlower
            else: wlist_left = None

            ww = (Rc >= wlower) & (Rc <= wupper)

        Dw = D[ww]
        Rw = Rc[ww]

        # Drop NA values

        if X is not None:
            Xw = X[ww, :]
            data = np.column_stack((Rw, Dw, Xw))
            data = data[~np.isnan(data).any(axis=1)]
            Rw = data[:, 0]
            Dw = data[:, 1]
            Xw = data[:, 2:]
        else:
            data = np.column_stack((Rw, Dw))
            data = data[~np.isnan(data).any(axis=1)]
            Rw = data[:, 0]
            Dw = data[:, 1]

        # Sample sizes

        n0_w = np.sum(Dw == 0)
        n1_w = np.sum(Dw == 1)
        n_w = n0_w + n1_w
        table_rdw[j, 3] = n0_w
        table_rdw[j, 4] = n1_w

        if n0_w == 0 or n1_w == 0:
            table_rdw[j, 0] = np.nan
            table_rdw[j, 1] = np.nan
            varname = ''
        else:

            # Binomial test

            p_value = binom_test(np.sum(Dw), len(Dw), p=0.5)
            table_rdw[j, 2] = p_value

            if X is not None:

                # Weights

                kweights = np.ones(n_w)

                if kernel == 'triangular':
                    kweights = (1 - np.abs(Rw / wupper)) * (np.abs(Rw / wupper) <= 1)
                    kweights[kweights == 0] = np.finfo(float).eps
                elif kernel == 'epan':
                    kweights = 0.75 * (1 - (Rw / wupper) ** 2) * (np.abs(Rw / wupper) <= 1)
                    kweights[kweights == 0] = np.finfo(float).eps

                # Model adjustment

                if p > 0:
                    X_adj = np.empty_like(Xw)

                    if evalat == 'cutoff':
                        evall = cutoff
                        evalr = cutoff
                    elif evalat == 'means':
                        evall = np.mean(Rw[Dw == 0]) + cutoff
                        evalr = np.mean(Rw[Dw == 1]) + cutoff

                    R_adj = Rw + cutoff - Dw * evalr - (1 - Dw) * evall
                    Rpoly = np.polynomial.polynomial.polyvander(R_adj, deg=p)

                    X_adj = np.zeros_like(Xw)

                    for k in range(Xw.shape[1]):
                        lfit_t = sm.WLS(Xw[Dw == 1, k], sm.add_constant(Rpoly[Dw == 1]), weights=kweights[Dw == 1]).fit()
                        X_adj[Dw == 1, k] = lfit_t.resid + lfit_t.params[0]

                        lfit_c = sm.WLS(Xw[Dw == 0, k], sm.add_constant(Rpoly[Dw == 0]), weights=kweights[Dw == 0]).fit()
                        X_adj[Dw == 0, k] = lfit_c.resid + lfit_c.params[0]

                        Xw = X_adj

                # Statistics and p-values
                if statistic == 'hotelling':
                    obs_stat = hotelT2(Xw, Dw)['statistic']
                    if not approx:
                        stat_distr = np.empty(reps)
                        for i in range(reps):
                            D_sample = np.random.choice(Dw, replace=False)
                            obs_stat_sample = hotelT2(Xw, D_sample).statistic
                            stat_distr[i] = obs_stat_sample
                        p_value = np.mean(np.abs(stat_distr) >= np.abs(obs_stat))
                    else:
                        p_value = hotelT2(Xw, Dw)['p.value']
                    table_rdw[j, 0] = p_value
                    varname = np.nan
                else:
                    result = rdrandinf_model(Xw, Dw, statistic=statistic, kweights=kweights, pvalue=True)
                    obs_stat = result['statistic']
                    if not approx:
                        stat_distr = np.empty((reps, X.shape[1]))
                        for i in range(reps):
                            D_sample = np.random.choice(Dw, size = len(Dw), replace=False)
                            obs_stat_sample = rdrandinf_model(Xw, D_sample, statistic=statistic, kweights=kweights)['statistic']
                            stat_distr[i, :] = obs_stat_sample
                        p_value = np.mean(np.abs(stat_distr) >= np.abs(obs_stat), axis=0)
                    else:
                        if p == 0:
                            p_value = result['p_value']
                        else:
                            p_value = np.zeros(X.shape[1])
                            for k in range(X.shape[1]):
                                lfit = sm.WLS(Xw[:, k], sm.add_constant(np.column_stack((Dw, Rpoly, Dw * Rpoly))), weights=kweights).fit()
                                tstat = lfit.params[1] / np.sqrt(lfit.cov_HC2.loc[1, 1])
                                p_value[k] = 2 * norm.cdf(-np.abs(tstat))

                    table_rdw[j, 0] = np.min(p_value)
                    tmp = np.argmin(p_value)
                    table_rdw[j, 1] = tmp
                    if colnames_X[tmp] is not None:
                        if colnames_X[tmp] != '':
                            varname = colnames_X[tmp]
                        else:
                            varname = tmp
                    else:
                        varname = tmp
            else:
                table_rdw[j, 0] = np.nan
                table_rdw[j, 1] = np.nan
                varname = np.nan

        table_rdw[j, 5] = wlower
        table_rdw[j, 6] = wupper

        if not quietly:
            print(f"{wlower+cutoff:9.4f}{wupper+cutoff:9.4f}{table_rdw[j, 0]:11.3f}{str(varname)[:15]:>16}{table_rdw[j, 2]:12.3f}{table_rdw[j, 3]:11.0f}{table_rdw[j, 4]:11.0f}")

    if not quietly: print('=' * 80)

###############################################################################
# Find recommended window
###############################################################################

    if X is not None:
        Pvals = table_rdw[:, 0]

        if (not np.isnan(Pvals[0])) and (Pvals[0] < level):
            print('Smallest window does not pass covariate test.')
            print('Decrease smallest window or reduce level.')
            tmp = -1
            rec_length = np.full((2,),np.nan)
            rec_window = np.full((2,),np.nan)
        elif np.all(Pvals >= level):
            tmp = len(Pvals)-1
            rec_window = [cutoff + table_rdw[tmp, 5], cutoff + table_rdw[tmp, 6]]
        else:
            tmp = np.min(np.where(Pvals < level))
            tmp -= 1
            rec_window = [cutoff + table_rdw[tmp, 5], cutoff + table_rdw[tmp, 6]]

        if (not quietly) and (tmp != -1):
            print(f"Recommended window is [{round(rec_window[0], 4)};{round(rec_window[1], 4)}] with "
                f"{table_rdw[tmp, 3] + table_rdw[tmp, 4]:.0f} observations "
                f"({table_rdw[tmp, 3]:.0f} below, {table_rdw[tmp, 4]:.0f} above).")
            print('\n\n')
    else:
        if not quietly:
            print('Note: no covariates specified.')
            print('Need to specify covariates to find recommended length.')
        rec_window = np.full((2,),np.nan)

    ###############################################################################
    # Plot p-values
    ###############################################################################

    if plot:
        if 'X' in locals():
            rdwinselect_plot = plt.figure().gca()
            rdwinselect_plot.set_xlabel('Windows Right')
            rdwinselect_plot.set_ylabel('p-values')
            rdwinselect_plot.scatter(wlist_right, Pvals)
            plt.show()
        else:
            raise ValueError('Cannot draw plot without covariates')

    ###############################################################################
    # Output
    ###############################################################################

    table_sumstats_index = ['Number of obs', '1th percentile', '5th percentile', '10th percentile', '20th percentile']
    table_sumstats_columns = ['Left of c', 'Right of c']
    table_sumstats = pd.DataFrame(table_sumstats, 
                         index = table_sumstats_index, 
                         columns = table_sumstats_columns)
    
    table_rdw_columns = ['p-value', 'Variable', 'Bi.test', 'Obs<c', 'Obs>=c', 'w_left', 'w_right']
    table_rdw = pd.DataFrame(table_rdw, 
                         columns = table_rdw_columns)
    output = {
        'w_left': rec_window[0],
        'w_right': rec_window[1],
        'wlist_left': wlist_left,
        'wlist_right': wlist_right,
        'results': table_rdw,
        'summary': table_sumstats
    }

    return output


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from rdlocrand.rdwinselect import rdwinselect
from rdlocrand.rdrandinf import rdrandinf

def rdrbounds(Y, R, cutoff=0, wlist=None, gamma=None, expgamma=None,
              bound='both', statistic='ranksum', p=0, evalat='cutoff',
              kernel='uniform', fuzzy=None, nulltau=0, prob=None,
              fmpval=False, reps=1000, seed=666):

    """
    Rosenbaum bounds for RD designs under local randomization

    rdrbounds calculates lower and upper bounds for the randomization p-value under different degrees of departure from a local randomized experiment, as suggested by Rosenbaum (2002).

    Author:
    Matias Cattaneo, Princeton University. Email: cattaneo@princeton.edu
    Rocio Titiunik, Princeton University. Email: titiunik@princeton.edu
    Gonzalo Vazquez-Bare, UC Santa Barbara. Email: gvazquez@econ.ucsb.edu

    References:
    Cattaneo, M.D., R. Titiunik, and G. Vazquez-Bare. (2016).
    Inference in Regression Discontinuity Designs under Local Randomization.
    Stata Journal 16(2): 331-367.
    URL: https://rdpackages.github.io/references/Cattaneo-Titiunik-VazquezBare_2016_Stata.pdf

    Rosenbaum, P. (2002). Observational Studies. Springer.

    Parameters:
    - Y: A vector containing the values of the outcome variable.
    - R: A vector containing the values of the running variable.
    - cutoff: The RD cutoff (default is 0).
    - wlist: The list of window lengths to be evaluated. By default, the program constructs 10 windows around the cutoff, the first one including 10 treated and control observations and adding 5 observations to each group in subsequent windows.
    - gamma: The list of values of gamma to be evaluated.
    - expgamma: The list of values of exp(gamma) to be evaluated. Default is [1.5, 2, 2.5, 3].
    - bound: Specifies which bounds the command calculates. Options are 'upper' for upper bound, 'lower' for lower bound, and 'both' for both upper and lower bounds. Default is 'both'.
    - statistic: The statistic to be used in the balance tests. Allowed options are 'diffmeans' (difference in means statistic), 'ksmirnov' (Kolmogorov-Smirnov statistic), and 'ranksum' (Wilcoxon-Mann-Whitney standardized statistic). Default option is 'diffmeans'. The statistic 'ttest' is equivalent to 'diffmeans' and included for backward compatibility.
    - p: The order of the polynomial for the outcome adjustment model. Default is 0.
    - evalat: Specifies the point at which the adjusted variable is evaluated. Allowed options are 'cutoff' and 'means'. Default is 'cutoff'.
    - kernel: Specifies the type of kernel to use as a weighting scheme. Allowed kernel types are 'uniform' (uniform kernel), 'triangular' (triangular kernel), and 'epan' (Epanechnikov kernel). Default is 'uniform'.
    - fuzzy: Indicates that the RD design is fuzzy. 'fuzzy' can be specified as a vector containing the values of the endogenous treatment variable or as a list where the first element is the vector of endogenous treatment values and the second element is a string containing the name of the statistic to be used. Allowed statistics are 'ar' (Anderson-Rubin statistic) and 'tsls' (2SLS statistic). Default statistic is 'ar'. The 'tsls' statistic relies on a large-sample approximation.
    - nulltau: The value of the treatment effect under the null hypothesis. Default is 0.
    - prob: The probabilities of treatment for each unit when the assignment mechanism is a Bernoulli trial. This option should be specified as a vector of length equal to the length of the outcome and running variables.
    - fmpval: Reports the p-value under fixed margins randomization, in addition to the p-value under Bernoulli trials.
    - reps: Number of replications. Default is 1000.
    - seed: The seed to be used for the randomization tests.

    Returns:
    - gamma: List of gamma values.
    - expgamma: List of exp(gamma) values.
    - wlist: Window grid.
    - p_values: p-values for each window (under gamma = 0).
    - lower_bound: List of lower bound p-values for each window and gamma pair.
    - upper_bound: List of upper bound p-values for each window and gamma pair.

    Examples:
    # Toy dataset
    import numpy as np

    R = np.random.uniform(-1, 1, size=100)
    Y = 1 + R - 0.5 * R**2 + 0.3 * R**3 + (R >= 0) + np.random.normal(size=100)

    # Rosenbaum bounds
    # Note: low number of replications and windows to speed up the process.
    # The user should increase these values.
    rdrbounds(Y, R, expgamma=[1.5, 2], wlist=[0.3], reps=100)

    """

    ###############################################################################
    # Parameters and error checking
    ###############################################################################

    if cutoff <= np.nanmin(R) or cutoff >= np.nanmax(R):
        raise ValueError('Cutoff must be within the range of the running variable')

    if bound != 'both' and bound != 'upper' and bound != 'lower':
        raise ValueError('bound option incorrectly specified')

    data = np.column_stack((Y, R))
    data = data[~np.isnan(data).any(axis=1)]
    Y = data[:, 0]
    R = data[:, 1]

    Rc = R - cutoff
    D = (Rc >= 0).astype(int)

    if gamma is None and expgamma is None:
        gammalist = [1.5, 2, 2.5, 3]
    elif gamma is None and expgamma is not None:
        gammalist = expgamma
    elif gamma is not None and expgamma is None:
        gammalist = np.exp(gamma)
    else:
        raise ValueError('gamma and expgamma cannot be specified simultaneously')

    if wlist is None:
        aux = rdwinselect(Rc, wobs=5, nwindows=5, quietly=True)
        wlist = np.round(aux['results'][:, 0], 2)

    if seed > 0:
        np.random.seed(seed)
    elif seed != -1:
        raise ValueError('Seed has to be a positive integer or -1 for system seed')

    evall = cutoff
    evalr = cutoff

    ###############################################################################
    # Randomization p-value
    ###############################################################################

    print('\nCalculating randomization p-value...\n')

    P = np.zeros((2, len(wlist)))

    count = 0

    if not fmpval:
        for w in wlist:
            ww = (np.round(Rc, 8) >= np.round(-w, 8)) & (np.round(Rc, 8) <= np.round(w, 8))
            Dw = D[ww]
            Rw = Rc[ww]

            if prob is None:
                prob_be = np.repeat(np.mean(Dw), len(R))
            else:
                prob_be = prob

            if evalat == 'means':
                evall = np.mean(Rw[Dw == 0])
                evalr = np.mean(Rw[Dw == 1])

            aux = rdrandinf(Y, Rc, wl=-w, wr=w, bernoulli=prob_be, reps=reps, p=p,
                            nulltau=nulltau, statistic=statistic,
                            evall=evall, evalr=evalr, kernel=kernel, fuzzy=fuzzy,
                            quietly=True)

            P[0, count] = aux['p.value']

            print('Bernoulli p-value (w = {}) = {}'.format(w, round(P[0, count], 3)))

            count += 1
    else:
        for w in wlist:
            ww = (np.round(Rc, 8) >= np.round(-w, 8)) & (np.round(Rc, 8) <= np.round(w, 8))
            Dw = D[ww]
            Rw = Rc[ww]

            if prob is None:
                prob_be = np.repeat(np.mean(Dw), len(R))
            else:
                prob_be = prob

            if evalat == 'means':
                evall = np.mean(Rw[Dw == 0])
                evalr = np.mean(Rw[Dw == 1])

            aux_be = rdrandinf(Y, Rc, wl=-w, wr=w, bernoulli=prob_be, reps=reps, p=p,
                               nulltau=nulltau, statistic=statistic,
                               evall=evall, evalr=evalr, kernel=kernel, fuzzy=fuzzy,
                               quietly=True)

            P[0, count] = aux_be['p.value']

            aux_fm = rdrandinf(Y, Rc, wl=-w, wr=w, reps=reps, p=p,
                               nulltau=nulltau, statistic=statistic,
                               evall=evall, evalr=evalr, kernel=kernel, fuzzy=fuzzy,
                               quietly=True)

            P[1, count] = aux_fm['p.value']

            print('Bernoulli p-value (w = {}) = {}'.format(w, round(P[0, count], 3)))
            print('Fixed margins p-value (w = {}) = {}'.format(w, round(P[1, count], 3)))

            count += 1

    print('\n')

    ###############################################################################
    # Sensitivity analysis
    ###############################################################################

    print('Running sensitivity analysis...\n')

    if bound == 'upper':
        p_ub = np.zeros((len(gammalist), len(wlist)))
        count_g = 0

        for G in gammalist:
            plow = 1 / (1 + G)
            phigh = G / (1 + G)
            count_w = 0

            for w in wlist:
                ww = (np.round(Rc, 8) >= np.round(-w, 8)) & (np.round(Rc, 8) <= np.round(w, 8))
                Dw = D[ww]
                Yw = Y[ww]
                Rw = Rc[ww]

                data_w = np.column_stack((Yw, Rw, Dw))
                jj = np.argsort(data_w[:, 0])[::-1]
                data_dec = data_w[jj, :]
                Yw_dec = data_dec[:, 0]
                Rw_dec = data_dec[:, 1]

                nw = len(Rw)
                nw1 = np.sum(Dw)
                nw0 = nw - nw1
                pvals_ub = []

                for u in range(1, nw + 1):
                    uplus = np.concatenate((np.ones(u), np.zeros(nw - u)))
                    p_aux = phigh * uplus + plow * (1 - uplus)
                    aux = rdrandinf(Yw_dec, Rw_dec, wl=-w, wr=w, bernoulli=p_aux, reps=reps, p=p,
                                    nulltau=nulltau, statistic=statistic,
                                    evall=evall, evalr=evalr, kernel=kernel, fuzzy=fuzzy,
                                    quietly=True)
                    pvals_ub.append(aux['p.value'])

                p_ub_w = max(pvals_ub)
                p_ub[count_g, count_w] = p_ub_w

                count_w += 1

            count_g += 1

    if bound == 'both':
        p_ub = np.zeros((len(gammalist), len(wlist)))
        p_lb = np.zeros((len(gammalist), len(wlist)))
        count_g = 0

        for G in gammalist:
            plow = 1 / (1 + G)
            phigh = G / (1 + G)
            count_w = 0

            for w in wlist:
                ww = (np.round(Rc, 8) >= np.round(-w, 8)) & (np.round(Rc, 8) <= np.round(w, 8))
                Dw = D[ww]
                Yw = Y[ww]
                Rw = Rc[ww]

                data_w = np.column_stack((Yw, Rw, Dw))
                ii = np.argsort(data_w[:, 0])
                data_inc = data_w[ii, :]
                Yw_inc = data_inc[:, 0]
                Rw_inc = data_inc[:, 1]
                jj = np.argsort(data_w[:, 0])[::-1]
                data_dec = data_w[jj, :]
                Yw_dec = data_dec[:, 0]
                Rw_dec = data_dec[:, 1]

                nw = len(Rw)
                nw1 = np.sum(Dw)
                nw0 = nw - nw1
                pvals_ub = []
                pvals_lb = []

                for u in range(1, nw + 1):
                    uplus = np.concatenate((np.ones(u), np.zeros(nw - u)))
                    p_aux = phigh * uplus + plow * (1 - uplus)
                    aux = rdrandinf(Yw_dec, Rw_dec, wl=-w, wr=w, bernoulli=p_aux, reps=reps, p=p,
                                    nulltau=nulltau, statistic=statistic,
                                    evall=evall, evalr=evalr, kernel=kernel, fuzzy=fuzzy,
                                    quietly=True)
                    pvals_ub.append(aux['p.value'])

                    uminus = np.concatenate((np.zeros(nw - u), np.ones(u)))
                    p_aux = phigh * uminus + plow * (1 - uminus)
                    aux = rdrandinf(Yw_inc, Rw_inc, wl=-w, wr=w, bernoulli=p_aux, reps=reps, p=p,
                                    nulltau=nulltau, statistic=statistic,
                                    evall=evall, evalr=evalr, kernel=kernel, fuzzy=fuzzy,
                                    quietly=True)
                    pvals_lb.append(aux['p.value'])

                p_ub_w = max(pvals_ub)
                p_lb_w = min(pvals_lb)
                p_ub[count_g, count_w] = p_ub_w
                p_lb[count_g, count_w] = p_lb_w

                count_w += 1

            count_g += 1

    if bound == 'lower':
        p_lb = np.zeros((len(gammalist), len(wlist)))
        count_g = 0

        for G in gammalist:
            plow = 1 / (1 + G)
            phigh = G / (1 + G)
            count_w = 0

            for w in wlist:
                ww = (np.round(Rc, 8) >= np.round(-w, 8)) & (np.round(Rc, 8) <= np.round(w, 8))
                Dw = D[ww]
                Yw = Y[ww]
                Rw = Rc[ww]

                data_w = np.column_stack((Yw, Rw, Dw))
                ii = np.argsort(data_w[:, 0])
                data_inc = data_w[ii, :]
                Yw_inc = data_inc[:, 0]
                Rw_inc = data_inc[:, 1]

                nw = len(Rw)
                nw1 = np.sum(Dw)
                nw0 = nw - nw1
                pvals_lb = []

                for u in range(1, nw + 1):
                    uminus = np.concatenate((np.zeros(nw - u), np.ones(u)))
                    p_aux = phigh * uminus + plow * (1 - uminus)
                    aux = rdrandinf(Yw_inc, Rw_inc, wl=-w, wr=w, bernoulli=p_aux, reps=reps, p=p,
                                    nulltau=nulltau, statistic=statistic,
                                    evall=evall, evalr=evalr, kernel=kernel, fuzzy=fuzzy,
                                    quietly=True)
                    pvals_lb.append(aux['p.value'])

                p_lb_w = min(pvals_lb)
                p_lb[count_g, count_w] = p_lb_w

                count_w += 1

            count_g += 1

    output = {}

    print('\nSensitivity analysis complete.')

    ###############################################################################
    # Output
    ###############################################################################

    if not fmpval:
        P = P[1:, :]

    if bound == 'both':
        output['gamma'] = np.log(gammalist)
        output['expgamma'] = gammalist
        output['wlist'] = wlist
        output['p.values'] = P
        output['lower.bound'] = p_lb
        output['upper.bound'] = p_ub

    if bound == 'upper':
        output['gamma'] = np.log(gammalist)
        output['expgamma'] = gammalist
        output['wlist'] = wlist
        output['p.values'] = P
        output['upper.bound'] = p_ub

    if bound == 'lower':
        output['gamma'] = np.log(gammalist)
        output['expgamma'] = gammalist
        output['wlist'] = wlist
        output['p.values'] = P
        output['lower.bound'] = p_lb

    return output


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from scipy.stats import norm
import statsmodels.api as sm
from linearmodels import IV2SLS
from rdlocrand.rdwinselect import rdwinselect
from rdlocrand.rdlocrand_fun import rdrandinf_model, find_CI

def rdrandinf(Y, R, cutoff=0, wl=None, wr=None, statistic='diffmeans', p=0, evall=None, evalr=None, kernel='uniform',
              fuzzy=None, nulltau=0, d=None, dscale=None, ci=None, interfci=None, bernoulli=None, reps=1000, seed=666,
              quietly=False, covariates=None, obsmin=None, wmin=None, wobs=None, wstep=None, wasymmetric=False,
              wmasspoints=False, nwindows=10, dropmissing=False, rdwstat='diffmeans', approx=False, rdwreps=1000,
              level=0.15, plot=False, firststage=False, obsstep=None):
    
    """
    Randomization Inference for RD Designs under Local Randomization

    rdrandinf implements randomization inference and related methods for RD designs,
    using observations in a specified or data-driven selected window around the cutoff where
    local randomization is assumed to hold.

    Parameters:
    -----------
    Y : array-like
        A vector containing the values of the outcome variable.
    R : array-like
        A vector containing the values of the running variable.
    cutoff : float, optional
        The RD cutoff (default is 0).
    wl : float, optional
        The left limit of the window. The default takes the minimum of the running variable.
    wr : float, optional
        The right limit of the window. The default takes the maximum of the running variable.
    statistic : str, optional
        The statistic to be used in the balance tests. Allowed options are 'diffmeans' (difference in means statistic),
        'ksmirnov' (Kolmogorov-Smirnov statistic), and 'ranksum' (Wilcoxon-Mann-Whitney standardized statistic).
        Default option is 'diffmeans'. The statistic 'ttest' is equivalent to 'diffmeans' and included for backward compatibility.
    p : int, optional
        The order of the polynomial for the outcome transformation model (default is 0).
    evall : float, optional
        The point at the left of the cutoff at which to evaluate the transformed outcome. Default is the cutoff value.
    evalr : float, optional
        The point at the right of the cutoff at which the transformed outcome is evaluated. Default is the cutoff value.
    kernel : str, optional
        Specifies the type of kernel to use as a weighting scheme. Allowed kernel types are 'uniform' (uniform kernel),
        'triangular' (triangular kernel), and 'epan' (Epanechnikov kernel). Default is 'uniform'.
    fuzzy : None or tuple or array-like, optional
        Indicates that the RD design is fuzzy. If fuzzy is None, the RD design is not fuzzy. If fuzzy is a tuple or array-like,
        the first element should be the vector of endogenous treatment values, and the second element should be a string
        containing the name of the statistic to be used. Allowed statistics are 'itt' (intention-to-treat statistic)
        and 'tsls' (2SLS statistic). Default statistic is 'ar'. The 'tsls' statistic relies on large-sample approximation.
    nulltau : float, optional
        The value of the treatment effect under the null hypothesis (default is 0).
    d : float, optional
        The effect size for asymptotic power calculation. Default is 0.5 * standard deviation of outcome variable for the control group.
    dscale : float, optional
        The fraction of the standard deviation of the outcome variable for the control group used as an alternative hypothesis
        for asymptotic power calculation. Default is 0.5.
    ci : float or array-like, optional
        Calculates a confidence interval for the treatment effect by test inversion. ci can be specified as a scalar or a vector,
        where the first element indicates the value of alpha for the confidence interval (typically 0.05 or 0.01),
        and the remaining elements, if specified, indicate the grid of treatment effects to be evaluated.
        This option uses rdsensitivity to calculate the confidence interval. See corresponding help for details.
        Note: the default tlist can be narrow in some cases, which may truncate the confidence interval.
        We recommend the user to manually set a large enough tlist.
    interfci : float, optional
        The level for Rosenbaum's confidence interval under arbitrary interference between units.
    bernoulli : array-like, optional
        The probabilities of treatment for each unit when the assignment mechanism is a Bernoulli trial.
        This option should be specified as a vector of length equal to the length of the outcome and running variables.
    reps : int, optional
        The number of replications (default is 1000).
    seed : int, optional
        The seed to be used for the randomization test.
    quietly : bool, optional
        Suppresses the output table.
    covariates : array-like, optional
        The covariates used by rdwinselect to choose the window when wl and wr are not specified.
        This should be a matrix of size n x k where n is the total sample size and k is the number of covariates.
    obsmin : int, optional
        The minimum number of observations above and below the cutoff in the smallest window employed by the companion command rdwinselect.
        Default is 10.
    wmin : float, optional
        The smallest window to be used (if obsmin is not specified) by the companion command rdwinselect.
        Specifying both wmin and obsmin returns an error.
    wobs : int, optional
        The number of observations to be added at each side of the cutoff at each step.
    wstep : float, optional
        The increment in window length (if obsstep is not specified) by the companion command rdwinselect.
        Specifying both obsstep and wstep returns an error.
    wasymmetric : bool, optional
        Allows for asymmetric windows around the cutoff when wobs is specified.
    wmasspoints : bool, optional
        Specifies that the running variable is discrete and each mass point should be used as a window.
    nwindows : int, optional
        The number of windows to be used by the companion command rdwinselect. Default is 10.
    dropmissing : bool, optional
        Drop rows with missing values in covariates when calculating windows.
    rdwstat : str, optional
        The statistic to be used by the companion command rdwinselect (see corresponding help for options).
        Default option is 'ttest'.
    approx : bool, optional
        Forces the companion command rdwinselect to conduct the covariate balance tests using a large-sample approximation
        instead of finite-sample exact randomization inference methods.
    rdwreps : int, optional
        The number of replications to be used by the companion command rdwinselect. Default is 1000.
    level : float, optional
        The minimum accepted value of the p-value from the covariate balance tests to be used by the companion command rdwinselect.
        Default is 0.15.
    plot : bool, optional
        Draws a scatter plot of the minimum p-value from the covariate balance test against window length implemented
        by the companion command rdwinselect.
    firststage : bool, optional
        Reports the results from the first step when using tsls.
    obsstep : int, optional
        The minimum number of observations to be added on each side of the cutoff for the sequence of fixed-increment nested windows.
        Default is 2. This option is deprecated and only included for backward compatibility.
    
    Returns
    -------
    sumstats : DataFrame
        Summary statistics.
    obs_stat : float or array-like
        Observed statistic(s).
    p_value : float or array-like
        Randomization p-value(s).
    asy_pvalue : float or array-like
        Asymptotic p-value(s).
    window : tuple
        Chosen window.
    ci : DataFrame, optional
        Confidence interval (only if ci option is specified).
    interf_ci : DataFrame, optional
        Confidence interval under interference (only if interfci is specified).

    Example
    ------- 

    X = array(rnorm(200),dim=c(100,2))
    R = X[1,] + X[2,] + rnorm(100)
    Y = 1 + R -.5*R^2 + .3*R^3 + (R>=0) + rnorm(100)

    # Randomization inference in window (-.75,.75)
    tmp = rdrandinf(Y, R, wl=-.75, wr=.75)

    # Randomization inference in window (-.75,.75), all statistics
    tmp = rdrandinf(Y, R, wl=-.75, wr=.75, statistic='all')

    # Randomization inference with window selection
    # Note: low number of replications to speed up the process.
    # The user should increase the number of replications.
    tmp = rdrandinf(Y, R, statistic='all', covariates=X, wmin=.5, wstep=.125, rdwreps=500)
    """

    randmech = 'fixed margins'
    Rc_long = R - cutoff

    if (fuzzy is not None) and (fuzzy !=''):
        statistic = ''
        if isinstance(fuzzy, list) and (len(fuzzy)==2):
            fuzzy_tr = np.array(fuzzy[0])
            if (fuzzy[1] == 'ar') or (fuzzy[1] == 'itt'):
                fuzzy_stat = 'ar'
            elif fuzzy[1] == 'tsls':
                fuzzy_stat = 'wald'
            else:
                raise ValueError('Invalid fuzzy statistic')
        else:
            fuzzy_stat = 'ar'
            fuzzy_tr = np.array(fuzzy)
    else:
        fuzzy_stat = ''

    if (fuzzy is None) or (fuzzy ==''):
        if bernoulli is None:
            data = np.column_stack((Y, R))
            data = data[~np.isnan(data).any(axis=1)]
            Y = data[:, 0]
            R = data[:, 1]
        else:
            data = np.column_stack((Y, R, bernoulli))
            data = data[~np.isnan(data).any(axis=1)]
            Y = data[:, 0]
            R = data[:, 1]
            bernoulli = data[:, 2]
    else:
        if bernoulli is None:
            data = np.column_stack((Y, R, fuzzy_tr))
            data = data[~np.isnan(data).any(axis=1)]
            Y = data[:, 0]
            R = data[:, 1]
            fuzzy_tr = data[:, 2]
        else:
            data = np.column_stack((Y, R, bernoulli, fuzzy_tr))
            data = data[~np.isnan(data).any(axis=1)]
            Y = data[:, 0]
            R = data[:, 1]
            bernoulli = data[:, 2]
            fuzzy_tr = data[:, 3]

    if cutoff < np.min(R) or cutoff > np.max(R):
        raise ValueError('Cutoff must be within the range of the running variable')

    if p < 0:
        raise ValueError('p must be a positive integer')

    if fuzzy is None:
        if statistic != 'diffmeans' and statistic != 'ttest' and statistic != 'ksmirnov' and statistic != 'ranksum' and statistic != 'all':
            raise ValueError('Invalid statistic')
    
    if kernel != 'uniform' and kernel != 'triangular' and kernel != 'epan':
        raise ValueError('Invalid kernel')
    
    if kernel != 'uniform' and evall is not None and evalr is not None:
        if evall != cutoff or evalr != cutoff:
            raise ValueError('Kernel only allowed when evall=evalr=cutoff')
    
    if kernel != 'uniform' and statistic != 'ttest' and statistic != 'diffmeans':
        raise ValueError('Kernel only allowed for diffmeans')

    if ci is not None:
        if ci[0] > 1 or ci[0] < 0:
            raise ValueError('ci must be in [0,1]')
    
    if interfci is not None:
        if interfci > 1 or interfci < 0:
            raise ValueError('interfci must be in [0,1]')
        if statistic != 'diffmeans' and statistic != 'ttest' and statistic != 'ksmirnov' and statistic != 'ranksum':
            raise ValueError('interfci only allowed with ttest, ksmirnov or ranksum')
    
    if bernoulli is not None:
        randmech = 'Bernoulli'
        if np.max(bernoulli) > 1 or np.min(bernoulli) < 0:
            raise ValueError('bernoulli probabilities must be in [0,1]')
        if len(bernoulli) != len(R):
            raise ValueError('bernoulli should have the same length as the running variable')
    
    if wl is not None and wr is not None:
        wselect = 'set by user'
        if wl >= wr:
            raise ValueError('wl has to be smaller than wr')
        if wl > cutoff or wr < cutoff:
            raise ValueError('window does not include cutoff')
    
    if wl is None and wr is not None:
        raise ValueError('wl not specified')
    
    if wl is not None and wr is None:
        raise ValueError('wr not specified')
    
    if evall is not None and evalr is None:
        raise ValueError('evalr not specified')
    
    if evall is None and evalr is not None:
        raise ValueError('evall not specified')
    
    if d is not None and dscale is not None:
        raise ValueError('Cannot specify both d and dscale')

    Rc = R - cutoff
    D = np.array(Rc >= 0, dtype=float)

    n = len(D)
    n1 = np.sum(D)
    n0 = n - n1

    if seed > 0:
        np.random.seed(seed)
    elif seed != -1:
        raise ValueError('Seed has to be a positive integer or -1 for system seed')
    
    ###############################################################################
    # Window selection
    ###############################################################################

    if wl is None and wr is None:
        if covariates is None:
            wl = np.min(R, axis=0, initial=np.inf)
            wr = np.max(R, axis=0, initial=-np.inf)
            wselect = 'run. var. range'
        else:
            wselect = 'rdwinselect'
            if not quietly:
                print('\nRunning rdwinselect...\n')
            rdwlength = rdwinselect(Rc_long, covariates, obsmin=obsmin, obsstep=obsstep, wmin=wmin, wstep=wstep, wobs=wobs,
                                wasymmetric=wasymmetric, wmasspoints=wmasspoints, dropmissing=dropmissing, nwindows=nwindows,
                                statistic=rdwstat, approx=approx, reps=rdwreps, plot=plot, level=level, seed=seed, quietly=True)
            wl = cutoff + rdwlength['w_left']
            wr = cutoff + rdwlength['w_right']
            if not quietly:
                print('\nrdwinselect complete.\n')

    if not quietly:
        print(f'\nSelected window = [{round(wl, 3)};{round(wr, 3)}] \n')

    if evall is not None and evalr is not None:
        if evall < wl or evalr > wr:
            raise ValueError('evall and evalr need to be inside window')

    ww = (np.round(R, 8) >= np.round(wl, 8)) & (np.round(R, 8) <= np.round(wr, 8))

    Yw = Y[ww]
    Rw = Rc[ww]
    Dw = D[ww]

    if (fuzzy is not None) and (fuzzy != ''):
        Tw = fuzzy_tr[ww]

    if bernoulli is None:
        data = np.column_stack((Yw, Rw, Dw))
        data = data[~np.isnan(data).any(axis=1)]
        Yw = data[:, 0]
        Rw = data[:, 1]
        Dw = data[:, 2]
    else:
        Bew = bernoulli[ww]
        data = np.column_stack((Yw, Rw, Dw, Bew))
        data = data[~np.isnan(data).any(axis=1)]
        Yw = data[:, 0]
        Rw = data[:, 1]
        Dw = data[:, 2]
        Bew = data[:, 3]

    n_w = len(Dw)
    n1_w = int(np.sum(Dw))
    n0_w = n_w - n1_w

    ###############################################################################
    # Summary statistics
    ###############################################################################

    sumstats = np.zeros((5, 2))
    sumstats[0, :] = [n0, n1]
    sumstats[1, :] = [n0_w, n1_w]
    mean0 = np.mean(Yw[Dw == 0], axis=0)
    mean1 = np.mean(Yw[Dw == 1], axis=0)
    sd0 = np.nanstd(Yw[Dw == 0], axis=0, ddof =1)
    sd1 = np.nanstd(Yw[Dw == 1], axis=0, ddof =1)
    sumstats[2, :] = [mean0, mean1]
    sumstats[3, :] = [sd0, sd1]
    sumstats[4, :] = [wl, wr]

    if d is None and dscale is None:
        delta = 0.5 * sd0
    if d is not None and dscale is None:
        delta = d
    if d is None and dscale is not None:
        delta = dscale * sd0

    ###############################################################################
    # Weights
    ###############################################################################

    kweights = np.ones(n_w)

    if kernel == 'triangular':
        bwt = wr - cutoff
        bwc = wl - cutoff
        kweights[Dw == 1] = (1 - np.abs(Rw[Dw == 1] / bwt)) * (np.abs(Rw[Dw == 1] / bwt) < 1)
        kweights[Dw == 0] = (1 - np.abs(Rw[Dw == 0] / bwc)) * (np.abs(Rw[Dw == 0] / bwc) < 1)
    elif kernel == 'epan':
        bwt = wr - cutoff
        bwc = wl - cutoff
        kweights[Dw == 1] = 0.75 * (1 - (Rw[Dw == 1] / bwt) ** 2) * (np.abs(Rw[Dw == 1] / bwt) < 1)
        kweights[Dw == 0] = 0.75 * (1 - (Rw[Dw == 0] / bwc) ** 2) * (np.abs(Rw[Dw == 0] / bwc) < 1)

    ###############################################################################
    # Outcome adjustment: model and null hypothesis
    ###############################################################################

    Y_adj = Yw.copy()

    if p > 0:
        if evall is None and evalr is None:
            evall = cutoff
            evalr = cutoff
        R_adj = Rw + cutoff - Dw * evalr - (1 - Dw) * evall
        Rpoly = np.transpose(np.vstack([R_adj**k for k in range(1,p+1)]))
        lfit_t = sm.WLS(Yw[Dw == 1], sm.add_constant(Rpoly[Dw == 1,:]), weights=kweights[Dw == 1]).fit()
        Y_adj[Dw == 1] = lfit_t.resid + lfit_t.params[0]
        lfit_c = sm.WLS(Yw[Dw == 0], sm.add_constant(Rpoly[Dw == 0,:]), weights=kweights[Dw == 0]).fit()
        Y_adj[Dw == 0] = lfit_c.resid + lfit_c.params[0]
    
    if (fuzzy is None) or (fuzzy == ''):
        Y_adj_null = Y_adj - nulltau * Dw
    else:
        Y_adj_null = Y_adj - nulltau * Tw

    ###############################################################################
    # Observed statistics and asymptotic p-values
    ###############################################################################

    if (fuzzy is None) or (fuzzy == ''):
        results = rdrandinf_model(Y_adj_null, Dw, statistic=statistic, pvalue=True, kweights=kweights, delta=delta)
    else:
        results = rdrandinf_model(Y_adj_null, Dw, statistic=fuzzy_stat, endogtr=Tw, pvalue=True, kweights=kweights, delta=delta)
    
    obs_stat = results['statistic']
    
    if p == 0:
        if fuzzy_stat == 'wald':
            firststagereg = sm.OLS(Tw, sm.add_constant(Dw)).fit()
            aux = IV2SLS(dependent = Yw, 
                            exog = None,
                            endog = sm.add_constant(Tw),
                            instruments = sm.add_constant(Dw),
                            weights = kweights).fit(cov_type = 'robust')
            obs_stat = aux.params[1]
            se = aux.std_errors[1]
            ci_lb = obs_stat - 1.96 * se
            ci_ub = obs_stat + 1.96 * se
            tstat = obs_stat / se
            asy_pval = 2 * norm.cdf(-np.abs(tstat))
            asy_power = 1 - norm.cdf(1.96 - delta / se) + norm.cdf(-1.96 - delta / se)
        else:
            asy_pval = results['p_value']
            asy_power = results['asy_power']
    else:
        if statistic == 'diffmeans' or statistic == 'ttest' or statistic == 'all':
            X_inter = sm.add_constant(np.column_stack((Dw.reshape(-1,1),Rpoly,Dw.reshape(-1,1)*Rpoly)))
            lfit = sm.WLS(Yw, X_inter, weights=kweights).fit()
            se = lfit.HC2_se[1]
            tstat = lfit.params[1] / se
            asy_pval = 2 * norm.cdf(-np.abs(tstat))
            asy_power = 1 - norm.cdf(1.96 - delta / se) + norm.cdf(-1.96 - delta / se)
        if statistic == 'ksmirnov' or statistic == 'ranksum':
            asy_pval = np.nan
            asy_power = np.nan
        if statistic == 'all':
            asy_pval = [float(asy_pval), np.nan, np.nan]
            asy_power = [float(asy_power), np.nan, np.nan]
        
        if fuzzy_stat == 'wald':
            inter = Rpoly * Dw
            firststagereg = sm.OLS(Tw, sm.add_constant(Dw)).fit()
            aux = IV2SLS(dependent = Yw, 
                            exog = sm.add_constant([Rpoly, inter]),
                            endog = Tw,
                            instruments = Dw,
                            weights = kweights).fit(cov_type = 'robust')
            obs_stat = aux.params[-1]
            se = aux.std_errors[-1]
            ci_lb = obs_stat - 1.96 * se
            ci_ub = obs_stat + 1.96 * se
            tstat = aux.params['Tw'] / se
            asy_pval = 2 * norm.cdf(-np.abs(tstat))
            asy_power = 1 - norm.cdf(1.96 - delta / se) + norm.cdf(-1.96 - delta / se)

    ###############################################################################
    # Randomization-based inference
    ###############################################################################

    if statistic == 'all': stats_distr = np.empty((reps, 3))
    else: stats_distr  = np.empty((reps, 1))
    
    if not quietly:
        print('')
        print('Running randomization-based test...')
    
    if fuzzy_stat != 'wald':
        if bernoulli is None:
            max_reps = np.math.comb(n_w, n1_w)
            reps = min(reps, max_reps)
            if max_reps < reps:
                print(f'Chosen no. of reps > total no. of permutations.\nreps set to {reps}.')
            
            for i in range(reps):
                D_sample = np.random.choice(Dw, size = len(Dw), replace=False)
                if (fuzzy is None) or (fuzzy ==''):
                    obs_stat_sample = rdrandinf_model(Y_adj_null, D_sample, statistic, kweights=kweights, delta=delta)['statistic']
                else:
                    obs_stat_sample = rdrandinf_model(Y_adj_null, D_sample, statistic=fuzzy_stat, endogtr=Tw, kweights=kweights, delta=delta)['statistic']
                stats_distr[i,:] = obs_stat_sample
        else:
            for i in range(reps):
                D_sample = np.random.uniform(0, 1, n_w) <= Bew
                if (np.mean(D_sample) == 1) or (np.mean(D_sample) == 0):
                    stats_distr[i] = np.nan # ignore cases where bernoulli assignment mechanism gives no treated or no controls
                else:
                    obs_stat_sample = float(rdrandinf_model(Y_adj_null, D_sample, statistic, kweights=kweights, delta=delta)['statistic'])
                    stats_distr[i] = obs_stat_sample

        if not quietly:
            print('Randomization-based test complete.')
        
        if statistic == 'all':
            p_value1 = np.mean(np.abs(stats_distr[:, 0]) >= np.abs(obs_stat[0]), axis=0)
            p_value2 = np.mean(np.abs(stats_distr[:, 1]) >= np.abs(obs_stat[1]), axis=0)
            p_value3 = np.mean(np.abs(stats_distr[:, 2]) >= np.abs(obs_stat[2]), axis=0)
            p_value = [p_value1, p_value2, p_value3]
        else:
            p_value = np.nanmean(np.abs(stats_distr) >= np.abs(obs_stat))

    else:
        p_value = np.nan

    ###############################################################################
    # Confidence interval
    ###############################################################################
        
    if ci is not None:
        ci_alpha = ci[0]
        if fuzzy_stat != 'wald':
            wr_c = wr - cutoff
            wl_c = wl - cutoff
            if not np.isscalar(ci):
                t_list = ci[1:]
                aux = rdsensitivity_inner(Y, Rc, p=p, wlist=wr_c, wlist_left=wl_c, tlist=t_list, fuzzy=fuzzy_stat, ci=[wl_c, wr_c], ci_alpha=ci_alpha, reps=reps, quietly=quietly, seed=seed)
            else:
                aux = rdsensitivity_inner(Y, Rc, p=p, wlist=wr_c, wlist_left=wl_c, fuzzy=fuzzy_stat, ci=[wl_c, wr_c], ci_alpha=ci_alpha, reps=reps, quietly=quietly, seed=seed)
            conf_int = aux['ci']
        else:
            conf_int = np.array([[ci_lb, ci_ub]])
        if np.any(np.isnan(conf_int)):
            print('Consider a larger tlist in ci() option.')

    ###############################################################################
    # Confidence interval under interference
    ###############################################################################
    
    if interfci is not None:
        p_low = interfci / 2
        p_high = 1 - interfci / 2
        qq = np.quantile(stats_distr, [p_low, p_high])
        interf_ci = np.array([obs_stat[0] - qq[1], obs_stat[0] - qq[0]])

    ###############################################################################
    # Output and display results
    ###############################################################################

    output = {}

    if ci is None and interfci is None:
        output['sumstats'] = sumstats
        output['obs.stat'] = obs_stat
        output['p.value'] = p_value
        output['asy.pvalue'] = asy_pval
        output['window'] = [wl, wr]

    if ci is not None and interfci is None:
        output['sumstats'] = sumstats
        output['obs.stat'] = obs_stat
        output['p.value'] = p_value
        output['asy.pvalue'] = asy_pval
        output['window'] = [wl, wr]
        output['ci'] = conf_int

    if ci is None and interfci is not None:
        output['sumstats'] = sumstats
        output['obs.stat'] = obs_stat
        output['p.value'] = p_value
        output['asy.pvalue'] = asy_pval
        output['window'] = [wl, wr]
        output['interf.ci'] = interf_ci

    if ci is not None and interfci is not None:
        output['sumstats'] = sumstats
        output['obs.stat'] = obs_stat
        output['p.value'] = p_value
        output['asy.pvalue'] = asy_pval
        output['window'] = [wl, wr]
        output['ci'] = conf_int
        output['interf.ci'] = interf_ci

    if not quietly:
        if statistic == 'diffmeans' or statistic == 'ttest':
            statdisp = 'Diff. in means'
        elif statistic == 'ksmirnov':
            statdisp = 'Kolmogorov-Smirnov'
        elif statistic == 'ranksum':
            statdisp = 'Rank sum z-stat'
        elif fuzzy_stat == 'ar':
            statdisp = 'ITT'
        elif fuzzy_stat == 'wald':
            statdisp = 'TSLS'

        print('\n')
        print(f'{"Number of obs =":18}{n:14.0f}')
        print(f'{"Order of poly =":18}{p:14.0f}')
        print(f'{"Kernel type =":18}{kernel:>14}')
        print(f'{"Reps =":18}{reps:14.0f}')
        print(f'{"Window =":18}{wselect:>14}')
        print(f'{"H0:     tau  =":18}{nulltau:14.3f}')
        print(f'{"Randomization =":18}{randmech:>14}')
        print('\n')

        print(f'{"Cutoff c = ":10}{cutoff:^9.3f}{"Left of c":>12}{"Right of c":>12}')
        print(f'{"Number of obs":19}{n0:12.0f}{n1:12.0f}')
        print(f'{"Eff. number of obs":19}{n0_w:12.0f}{n1_w:12.0f}')
        print(f'{"Mean of outcome":19}{mean0:12.3f}{mean1:12.3f}')
        print(f'{"S.d. of outcome":19}{sd0:12.3f}{sd1:12.3f}')
        print(f'{"Window":19}{wl:12.3f}{wr:12.3f}')
      
        print('\n' + '=' * 80)

        if firststage and fuzzy_stat == 'wald':
            print("First stage regression")
            print(firststagereg.summary())
            print('\n' + '=' * 80 )

        print(f'{"":31}{"Finite sample":^20}{"Large sample":^29}')
        print(f'{"":31}{"-" * 18:18}{"":2}{"-" * 29:29}')
        print(f'{"Statistic":19}{"T":>11}{"P>|T|":^21}{"P>|T|":^9}{"Power vs d = ":>15}{delta:4.3f}')
       
        print('=' * 80)

        if statistic != 'all':
            if not np.isscalar(asy_pval): asy_pval = asy_pval[0]
            if not np.isscalar(asy_power): asy_power = asy_power[0]
            if not np.isscalar(obs_stat): obs_stat = obs_stat[0]
            print(f'{statdisp:19}{obs_stat:11.3f}{p_value:^21.3f}{asy_pval:^9.3f}{asy_power:20.3f}')

        if statistic == 'all':
            print(f"{'Diff. in means':19}{obs_stat[0]:11.3f}{p_value[0]:^21.3f}{asy_pval[0]:<9.3f}{asy_power[0]:>20.3f}")
            print(f"{'Kolmogorov-Smirnov':19}{obs_stat[1]:>11.3f}{p_value[1]:^21.3f}{asy_pval[1]:<9.3f}{asy_power[1]:>20.3f}")
            print(f"{'Rank sum z-stat':19}{obs_stat[2]:>11.3f}{p_value[2]:^21.3f}{asy_pval[2]:<9.3f}{asy_power[2]:>20.3f}")

        print('=' * 80)
        if ci is not None:
            print()
            if fuzzy_stat != 'wald':
                if len(conf_int) == 1:
                    print(f"{(1 - ci_alpha) * 100:.0f}% confidence interval: [{round(conf_int[0,0],3)},{round(conf_int[0,1],3)}]")
                else:
                    print(f"{(1 - ci_alpha) * 100:.0f}% confidence interval:")
                    print(np.round(conf_int, 3))
                    print()
                    print("Note: CI is disconnected - each row is a subset of the CI")
            else:
                print(f"{(1 - ci_alpha) * 100:.0f}% confidence interval: [{round(ci_lb, 3):.3f}, {round(ci_ub, 3):.3f}]")
                print("CI based on asymptotic approximation")

        if interfci is not None:
            print()
            print(f"{(1 - interfci) * 100:.0f}% confidence interval under interference: [{round(interf_ci[0], 3):.3f}, {round(interf_ci[1], 3):.3f}]")

    return output










def rdsensitivity_inner(Y, R, cutoff=0, wlist=None, wlist_left=None,
                   tlist=None, statistic='diffmeans', p=0,
                    evalat='cutoff', kernel='uniform', fuzzy=None,
                    ci=None, ci_alpha=0.05, reps=1000, seed=666, quietly=False):
    
    """
    This function is a copy of rdsensitivity to be called inside inside rdlocrand
    and avoid the circular reference when imporing the modules
    """
    
    ###############################################################################
    # Parameters and error checking
    ###############################################################################
    
    if cutoff < np.min(R) or cutoff > np.max(R):
        raise ValueError('Cutoff must be within the range of the running variable')
    if statistic not in ['diffmeans', 'ttest', 'ksmirnov', 'ranksum']:
        raise ValueError(statistic + ' not a valid statistic')
    if evalat not in ['cutoff', 'means']:
        raise ValueError('evalat only admits means or cutoff')
    if wlist_left is not None:
        if np.isscalar(wlist_left): wlist_left = np.array([wlist_left])
        if wlist is None:
            raise ValueError('Need to specify wlist when wlist_left is specified')
        elif np.isscalar(wlist): wlist= np.array([wlist])
        if len(wlist) != len(wlist_left):
            raise ValueError('Lengths of wlist and wlist_left need to coincide')
    if ci is not None and len(ci) != 2:
        raise ValueError('Need to specify wleft and wright in CI option')

    if seed > 0:
        np.random.seed(seed)
    elif seed != -1:
        raise ValueError('Seed has to be a positive integer or -1 for system seed')

    data = np.column_stack((Y, R))
    data = data[~np.isnan(data).any(axis=1)]
    Y = data[:, 0]
    R = data[:, 1]

    Rc = R - cutoff

    ###############################################################################
    # Default window list
    ###############################################################################

    if wlist is None:
        aux = rdwinselect(Rc, wobs=5, quietly=True)
        wlist = aux['results'][:, 6]
        wlist_left = aux['results'][:, 5]
    else:
        wlist_orig = wlist
        wlist = wlist - cutoff
        if wlist_left is None:
            wlist_left = -wlist
            wlist_left_orig = wlist_left
        else:
            wlist_left_orig = wlist_left
            wlist_left = wlist_left - cutoff

    wnum = len(wlist)
    

    ###############################################################################
    # Default tau list
    ###############################################################################

    if tlist is None:
        D = (Rc >= 0).astype(int)
        wfirst = max(wlist[0], abs(wlist_left[0]))
        if (fuzzy is None) or (fuzzy==''):
            Yaux = Y[np.abs(Rc) <= wfirst]
            Daux = D[np.abs(Rc) <= wfirst]
            model = sm.OLS(Yaux, sm.add_constant(Daux))
            results = model.fit()
            ci_lb = round(results.params[1] - 1.96 * np.sqrt(results.cov_params()[1, 1]), 2)
            ci_ub = round(results.params[1] + 1.96 * np.sqrt(results.cov_params()[1, 1]), 2)
        else:
            Yaux = Y[np.abs(Rc) <= wfirst]
            Daux = D[np.abs(Rc) <= wfirst]
            Taux = fuzzy[np.abs(Rc) <= wfirst]
            model = IV2SLS(dependent = Yaux, 
                            exog = None,
                            endog = sm.add_constant(Taux),
                            instruments = sm.add_constant(Daux))
            instrument_results = model.fit(cov_type = 'robust')
            ci_lb = round(instrument_results.params[1] - 1.96 * aux.std_errors[1], 2)
            ci_ub = round(instrument_results.params[1] + 1.96 * aux.std_errors[1], 2)

        wstep = round((ci_ub - ci_lb) / 10, 2)
        tlist = np.arange(ci_lb, ci_ub + wstep, wstep)

    ###############################################################################
    # Sensitivity analysis
    ###############################################################################

    if np.isscalar(tlist): tlist = np.array([tlist])

    results = np.empty((len(tlist), len(wlist)))
    if not quietly:
        print('')
        print('Running sensitivity analysis...', end="")

    for row, t in enumerate(tlist):
        for w in range(wnum):
            wright = wlist[w]
            wleft = wlist_left[w]
            if evalat == 'means':
                ww = (np.round(Rc, 8) >= np.round(wleft, 8)) & (np.round(Rc, 8) <= np.round(wright, 8))
                Rw = R[ww]
                Dw = D[ww]
                evall = np.mean(Rw[Dw == 0])
                evalr = np.mean(Rw[Dw == 1])
            else:
                evall = None
                evalr = None

            aux = rdrandinf(Y, Rc, wl=wleft, wr=wright, p=p, reps=reps, nulltau=t,
                               statistic=statistic, kernel=kernel, evall=evall, evalr=evalr,
                               fuzzy=fuzzy, seed=seed, quietly=True)
            results[row, w] = aux['p.value']

    if not quietly:
        print('Sensitivity analysis complete.\n')

    ###############################################################################
    # Confidence interval
    ###############################################################################

    conf_int = None
    if ci is not None:
        ci_window_l = ci[0] - cutoff
        ci_window_r = ci[1] - cutoff

        if np.isin(ci_window_r, wlist) and np.isin(ci_window_l, wlist_left):
            col = np.where(wlist == ci_window_r)[0][0]
            aux = results[:, col]

            conf_int = find_CI(aux, ci_alpha, tlist)
        else:
            raise ValueError('Window specified in ci not in wlist')
        
    ###############################################################################
    # Output
    ###############################################################################

    output = {'tlist': tlist, 'wlist': wlist_orig, 'wlist_left': wlist_left_orig, 'results': results}
    
    if conf_int is not None:
        output['ci'] = conf_int

    return output